"""Format typehints elegantly and and fix automatically created links.

The Sphinx extension :mod:`sphinx_autodoc_typehints` adds type annotations to functions.
This extension modifies the created type annotations in four ways:

#. It formats the annotations more simply and in line with e.g. :mod:`numpy`.
#. It defines a configuration value ``qualname_overrides`` for ``conf.py``
   that overrides automatically created links. It is used like this::

       qualname_overrides = {
           "pandas.core.frame.DataFrame": "pandas.DataFrame",
           ...,
       }

   The defaults include :class:`anndata.AnnData`, :class:`pandas.DataFrame`,
   :class:`scipy.sparse.spmatrix` and other classes in :mod:`scipy.sparse`.

   It is necessary since :attr:`~definition.__qualname__` does not necessarily match
   the documented location of the function/class.

   Once either `sphinx issue 4826`_ or `sphinx-autodoc-typehints issue 38`_ are fixed,
   this part of the functionality will no longer be necessary.
#. The config value ``annotate_defaults`` (default: :data:`True`) controls if rST code
   like ``(default: `42`)`` is added after the type.
   It sets sphinx-autodoc-typehints’s option ``typehints_defaults`` to ``'braces'``
#. Type annotations for :class:`tuple` return types are added::

       def x() -> Tuple[int, float]:
           \"""
           Returns:
               a: An integer
               b: A floating point number
           \"""

   will render as:

       :Returns: a : :class:`int`
                     An integer
                 b : :class:`float`
                     A floating point number


.. _sphinx issue 4826: https://github.com/sphinx-doc/sphinx/issues/4826
.. _sphinx-autodoc-typehints issue 38: https://github.com/tox-dev/sphinx-autodoc-typehints/issues/38

"""  # noqa: D300, D301

from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path
from collections import ChainMap
from dataclasses import dataclass

from sphinx.ext.autodoc import ClassDocumenter
from sphinx.ext.napoleon import NumpyDocstring  # type: ignore[attr-defined]

from scanpydoc import metadata, _setup_sig

from .example import example_func_prose, example_func_tuple


if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Callable

    from sphinx.config import Config
    from sphinx.application import Sphinx


__all__ = ["example_func_prose", "example_func_tuple", "setup"]


HERE = Path(__file__).parent.resolve()

qualname_overrides_default = {
    "anndata.base.AnnData": "anndata.AnnData",
    "anndata.core.anndata.AnnData": "anndata.AnnData",
    "anndata._core.anndata.AnnData": "anndata.AnnData",
    "matplotlib.axes._axes.Axes": "matplotlib.axes.Axes",
    "pandas.core.frame.DataFrame": "pandas.DataFrame",
    "pandas.core.indexes.base.Index": "pandas.Index",
    "scipy.sparse.base.spmatrix": "scipy.sparse.spmatrix",
    "scipy.sparse.csr.csr_matrix": "scipy.sparse.csr_matrix",
    "scipy.sparse.csc.csc_matrix": "scipy.sparse.csc_matrix",
}
qualname_overrides = ChainMap({}, qualname_overrides_default)


def _init_vars(_app: Sphinx, config: Config) -> None:
    qualname_overrides.update(config.qualname_overrides)
    if "sphinx_autodoc_typehints" not in config.extensions and config.annotate_defaults:
        msg = (
            "Can only use annotate_defaults option when using sphinx-autodoc-typehints"
        )
        raise ValueError(msg)
    if config.typehints_defaults is None and config.annotate_defaults:
        # override default for “typehints_defaults”
        config.typehints_defaults = "braces"  # type: ignore[attr-defined]


@dataclass
class PickleableCallable:
    func: Callable[..., Any]

    __call__ = property(lambda self: self.func)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Patches :mod:`sphinx_autodoc_typehints` for a more elegant display."""
    app.add_config_value("qualname_overrides", default={}, rebuild="html")
    app.add_config_value("annotate_defaults", default=True, rebuild="html")
    app.connect("config-inited", _init_vars)

    from ._formatting import typehints_formatter

    app.config["typehints_formatter"] = PickleableCallable(typehints_formatter)

    from ._autodoc_patch import dir_head_adder

    ClassDocumenter.add_directive_header = dir_head_adder(  # type: ignore[method-assign,assignment]
        qualname_overrides,
        ClassDocumenter.add_directive_header,
    )

    from ._return_tuple import process_docstring  # , process_signature
    from ._return_patch_numpydoc import _parse_returns_section

    NumpyDocstring._parse_returns_section = _parse_returns_section  # type: ignore[method-assign,assignment]  # noqa: SLF001
    app.connect("autodoc-process-docstring", process_docstring)

    return metadata



"""GitHub URLs for class and method pages.

This extension does two things:

#. It registers a :ref:`Jinja filter <jinja:filters>` called :func:`github_url`
   that you can use to convert a module path into a GitHub URL.
#. It configures :mod:`sphinx.ext.linkcode` for you if loaded after it in ``conf.py``:

   .. code:: python

      extensions = [
          "scanpydoc",
          "sphinx.ext.linkcode",
      ]

      # no need to define `linkcode_resolve`

Configuration
-------------

Uses the following config values in ``conf.py``::

    rtd_links_prefix: os.PathLike | str = ...  # default: '.'

    # sphinx book theme style
    html_theme_options = dict(
        repository_url=...,
        repository_branch=...,
    )
    # or RTD theme style:
    html_context = dict(
        github_user=...,
        github_repo=...,
        github_version=...,
    )

The ``rtd_links_prefix`` is for figuring out the .py file path relative to the git root,
that is to construct the path in the GitHub URL.

Which html configuration style you want to use depends on your theme, e.g.
:doc:`Sphinx Book Theme <sphinx_book_theme:index>`.

``:github_url:`` usage
----------------------

You can use the filter e.g. in `autosummary templates`_.
To configure the :doc:`Sphinx Book Theme <sphinx_book_theme:index>`,
override the ``autosummary/base.rst`` template like this:

.. code:: restructuredtext

    :github_url: {{ fullname | github_url }}

    {% extends "!autosummary/base.rst" %}

.. _autosummary templates: \
   http://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html#customizing-templates
"""
from __future__ import annotations

import sys
import inspect
from types import ModuleType
from typing import TYPE_CHECKING
from pathlib import Path, PurePosixPath
from importlib import import_module
from dataclasses import fields, is_dataclass

from jinja2.defaults import DEFAULT_FILTERS  # type: ignore[attr-defined]

from scanpydoc import metadata, _setup_sig


if TYPE_CHECKING:
    from types import CodeType, FrameType, MethodType, FunctionType, TracebackType
    from typing import Any, TypeAlias
    from collections.abc import Callable

    _SourceObjectType: TypeAlias = (
        ModuleType
        | type[Any]
        | MethodType
        | FunctionType
        | TracebackType
        | FrameType
        | CodeType
        | Callable[..., Any]
    )

    from sphinx.config import Config
    from sphinx.application import Sphinx


rtd_links_prefix: PurePosixPath | None = None
github_base_url: str | None = None


def _init_vars(_app: Sphinx, config: Config) -> None:  # pragma: no cover
    """Run hook when ``conf.py`` has been loaded."""
    global github_base_url, rtd_links_prefix
    github_base_url, rtd_links_prefix = _infer_vars(config)


def _infer_vars(config: Config) -> tuple[str, PurePosixPath]:
    _check_html_config(config)
    try:
        github_base_url = "https://github.com/{github_user}/{github_repo}/tree/{github_version}".format_map(
            config.html_context,
        )
    except (AttributeError, KeyError):
        github_base_url = "{repository_url}/tree/{repository_branch}".format_map(
            config.html_theme_options,
        )
    rtd_links_prefix = PurePosixPath(config.rtd_links_prefix)
    return github_base_url, rtd_links_prefix


def _get_annotations(obj: _SourceObjectType) -> dict[str, Any]:
    if sys.version_info > (3, 10):
        from inspect import get_annotations
    else:  # pragma: no cover
        from get_annotations import get_annotations

    try:
        return get_annotations(obj)  # type: ignore[arg-type]
    except TypeError:  # pragma: no cover
        return {}


def _get_obj_module(qualname: str) -> tuple[Any, ModuleType]:  # noqa: PLR0912
    """Get a module/class/attribute and its original module by qualname.

    Returns `None` as `obj` if it’s an annotated field without value.
    """
    modname = qualname
    attr_path: list[str] = []
    while "." in modname and modname not in sys.modules:
        modname, leaf = modname.rsplit(".", 1)
        attr_path.insert(0, leaf)

    # retrieve object and find original module name
    mod = sys.modules[modname]
    obj = None if attr_path else mod
    del modname
    for attr_name in attr_path:
        try:
            thing = getattr(mod if obj is None else obj, attr_name)
        except AttributeError as e:
            if is_dataclass(obj):
                thing = next(f for f in fields(obj) if f.name == attr_name)
            elif obj is not None and attr_name in _get_annotations(obj):
                thing = None
            else:  # pragma: no cover
                try:
                    thing = import_module(f"{mod.__name__}.{attr_name}")
                except ImportError:
                    raise e from None
        if isinstance(thing, ModuleType):  # pragma: no cover
            mod = thing
        elif is_dataclass(obj):
            obj = thing
        else:
            obj = thing
            mod_orig = getattr(obj, "__module__", None)
            if mod_orig is not None:
                mod = sys.modules[mod_orig]

    return obj, mod


def _get_linenos(obj: _SourceObjectType) -> tuple[int, int] | tuple[None, None]:
    """Get an object’s line numbers."""
    try:
        lines, start = inspect.getsourcelines(obj)
    except TypeError:
        return None, None
    else:
        return start, start + len(lines) - 1


def _module_path(obj: _SourceObjectType, module: ModuleType) -> PurePosixPath:
    """Relative module path to parent directory of toplevel module."""
    while hasattr(obj, "__wrapped__"):
        obj = obj.__wrapped__
    try:
        file = Path(inspect.getabsfile(obj))
    except TypeError:
        file = Path(module.__file__ or "")
    offset = -1 if file.name == "__init__.py" else 0
    parts = module.__name__.split(".")
    return PurePosixPath(*file.parts[offset - len(parts) :])


def github_url(qualname: str) -> str:
    """Get the full GitHub URL for some object’s qualname.

    Parameters
    ----------
    qualname
        The full qualified name of a function, class, method or module

    Returns
    -------
    A GitHub URL derived from the :confval:`html_context`
    or the :confval:`html_theme_options`.
    """
    try:
        obj, module = _get_obj_module(qualname)
    except Exception as e:
        if sys.version_info >= (3, 11):
            e.add_note(f"Qualname: {qualname!r}")
        raise
    assert rtd_links_prefix is not None  # noqa: S101
    path = rtd_links_prefix / _module_path(obj, module)
    start, end = _get_linenos(obj)
    fragment = f"#L{start}-L{end}" if start and end else ""
    return f"{github_base_url}/{path}{fragment}"


def _check_html_config(config: Config) -> None:
    options = dict(
        html_context={"github_user", "github_repo", "github_version"},
        html_theme_options={"repository_url", "repository_branch"},
    )
    if not any(name in config for name in options):
        msg = (
            f"Extension {__name__} needs “html_context” "
            "or “html_theme_options” to be defined in conf.py"
        )
        raise ValueError(msg)
    missing_value_sets = {
        name: opts - getattr(config, name, {}).keys() for name, opts in options.items()
    }
    if not all(missing_value_sets.values()):
        return  # a valid configuration was found

    mvs = " or ".join(
        f"{name} {', '.join(repr(mv) for mv in mvs)}"
        for name, mvs in missing_value_sets.items()
    )
    msg = f"Extension {__name__} needs {mvs} to be defined in conf.py."
    for name in options:
        if name in config:
            msg += f"\n{name} = {config[name]!r}"
    raise ValueError(msg)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Register the :func:`github_url` :ref:`Jinja filter <jinja:filters>`."""
    app.add_config_value("rtd_links_prefix", PurePosixPath("."), "")
    app.connect("config-inited", _init_vars)

    # if linkcode config not set
    if "linkcode_resolve" not in app.config or app.config["linkcode_resolve"] is None:
        from ._linkcode import linkcode_resolve

        app.config["linkcode_resolve"] = linkcode_resolve

    # html_context doesn’t apply to autosummary templates ☹
    # and there’s no way to insert filters into those templates
    # so we have to modify the default filters
    DEFAULT_FILTERS["github_url"] = github_url

    return metadata



from __future__ import annotations

import sys
import inspect
from typing import TYPE_CHECKING


if sys.version_info >= (3, 10):
    from types import UnionType
else:  # pragma: no cover
    UnionType = None


from scanpydoc import elegant_typehints


if TYPE_CHECKING:
    from typing import Any

    from sphinx.config import Config


def typehints_formatter(annotation: type[Any], config: Config) -> str | None:
    """Generate reStructuredText containing links to the types.

    Can be used as ``typehints_formatter`` for :mod:`sphinx_autodoc_typehints`,
    to respect the ``qualname_overrides`` option.

    Arguments
    ---------
    annotation
        A type or class used as type annotation.
    config
        Sphinx config containing ``sphinx-autodoc-typehints``’s options.

    Returns
    -------
    reStructuredText describing the type
    """
    if inspect.isclass(annotation) and annotation.__module__ == "builtins":
        return None

    tilde = "" if config.typehints_fully_qualified else "~"

    annotation_cls = annotation if inspect.isclass(annotation) else type(annotation)
    if annotation_cls.__module__ in {"typing", "types"}:
        return None

    # Only if this is a real class we override sphinx_autodoc_typehints
    if inspect.isclass(annotation):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = elegant_typehints.qualname_overrides.get(full_name)
        role = "exc" if issubclass(annotation_cls, BaseException) else "class"
        if override is not None:
            return f":py:{role}:`{tilde}{override}`"

    return None



"""Custom pytest fixtures and setup."""

from __future__ import annotations

import sys
import linecache
import importlib.util
from uuid import uuid4
from typing import TYPE_CHECKING
from textwrap import dedent

import pytest


if TYPE_CHECKING:
    from types import ModuleType
    from typing import Any
    from pathlib import Path
    from collections.abc import Callable, Generator

    from sphinx.application import Sphinx


@pytest.fixture()
def make_app_setup(
    make_app: Callable[..., Sphinx], tmp_path: Path
) -> Callable[..., Sphinx]:
    def make_app_setup(**conf: Any) -> Sphinx:  # noqa: ANN401
        (tmp_path / "conf.py").write_text("")
        return make_app(srcdir=tmp_path, confoverrides=conf)

    return make_app_setup


@pytest.fixture()
def make_module(
    tmp_path: Path,
) -> Generator[Callable[[str, str], ModuleType], None, None]:
    added_modules = []

    def make_module(name: str, code: str) -> ModuleType:
        code = dedent(code)
        assert name not in sys.modules
        spec = importlib.util.spec_from_loader(name, loader=None)
        assert spec is not None
        mod = sys.modules[name] = importlib.util.module_from_spec(spec)
        path = tmp_path / f"{name}_{str(uuid4()).replace('-', '_')}.py"
        path.write_text(code)
        mod.__file__ = str(path)
        exec(code, mod.__dict__)  # noqa: S102
        linecache.updatecache(str(path), mod.__dict__)
        added_modules.append(name)
        return mod

    yield make_module

    for name in added_modules:
        del sys.modules[name]



"""Test elegant_typehints subextension."""

from __future__ import annotations

import re
import sys
import inspect
from io import StringIO
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    AnyStr,
    NoReturn,
    Optional,
    cast,
    get_origin,
)
from pathlib import Path
from collections.abc import Mapping, Callable

import pytest
import sphinx_autodoc_typehints as sat
from sphinx.ext import napoleon
from sphinx.errors import ExtensionError
from sphinx_autodoc_typehints.patches import (
    napoleon_numpy_docstring_return_type_processor,
)

from scanpydoc.elegant_typehints._formatting import typehints_formatter
from scanpydoc.elegant_typehints._return_tuple import process_docstring


if TYPE_CHECKING:
    from types import ModuleType
    from typing import Protocol

    from sphinx.application import Sphinx

    class ProcessDoc(Protocol):  # noqa: D101
        def __call__(  # noqa: D102
            self, fn: Callable[..., Any], *, run_napoleon: bool = False
        ) -> list[str]:
            ...


NONE_RTYPE = ":rtype: :sphinx_autodoc_typehints_type:`\\:py\\:obj\\:\\`None\\``"


@pytest.fixture()
def testmod(make_module: Callable[[str, str], ModuleType]) -> ModuleType:
    return make_module(
        "testmod",
        """\
        class Class: pass
        class SubCl(Class): pass
        class Excep(RuntimeError): pass
        class Excep2(Excep): pass
        """,
    )


@pytest.fixture()
def app(make_app_setup: Callable[..., Sphinx]) -> Sphinx:
    return make_app_setup(
        master_doc="index",
        extensions=[
            "sphinx.ext.autodoc",
            "sphinx.ext.napoleon",
            "sphinx_autodoc_typehints",
            "scanpydoc.elegant_typehints",
        ],
        qualname_overrides={
            "testmod.Class": "test.Class",
            "testmod.SubCl": "test.SubCl",
            "testmod.Excep": "test.Excep",
            "testmod.Excep2": "test.Excep2",
        },
    )


@pytest.fixture()
def process_doc(app: Sphinx) -> ProcessDoc:
    def process(fn: Callable[..., Any], *, run_napoleon: bool = False) -> list[str]:
        lines = (inspect.getdoc(fn) or "").split("\n")
        if isinstance(fn, property):
            name = fn.fget.__name__
        elif hasattr(fn, "__name__"):
            name = fn.__name__
        else:
            name = "???"
        napoleon_numpy_docstring_return_type_processor(
            app, "function", name, fn, None, lines
        )
        if run_napoleon:
            napoleon._process_docstring(app, "function", name, fn, None, lines)  # noqa: SLF001
        sat.process_docstring(app, "function", name, fn, None, lines)
        process_docstring(app, "function", name, fn, None, lines)
        return lines

    return process


def test_app(app: Sphinx) -> None:
    assert "qualname_overrides" in app.config.values
    assert "testmod.Class" in app.config.qualname_overrides


def test_default(app: Sphinx) -> None:
    assert typehints_formatter(str, app.config) is None


def _escape_sat(rst: str) -> str:
    rst = (
        rst.replace("\\", r"\\")
        .replace("`", r"\`")
        .replace(":", r"\:")
        .replace("~", r"\~")
        .replace(",", r"\,")
        .replace("[", r"\[")
        .replace("]", r"\]")
    )
    return f":sphinx_autodoc_typehints_type:`{rst}`"


@pytest.mark.parametrize(
    ("kind", "add_rtype"),
    [
        pytest.param(lambda f: f, True, id="function"),
        pytest.param(property, False, id="property"),
    ],
)
def test_kinds(
    *,
    process_doc: ProcessDoc,
    kind: Callable[[Callable[..., Any]], Callable[..., Any]],
    add_rtype: bool,
) -> None:
    def fn_test(s: str) -> None:  # pragma: no cover
        """:param s: Test"""
        del s

    assert process_doc(kind(fn_test)) == [
        f":type s: {_escape_sat(':py:class:`str`')}",
        ":param s: Test",
        *([NONE_RTYPE] if add_rtype else []),
    ]


class CustomCls:  # noqa: D101
    __slots__ = ["foo"]

    def meth(self):  # pragma: no cover  # noqa: ANN201
        """No return section and no return annotation."""


@pytest.mark.parametrize(
    "obj",
    [
        pytest.param(None, id="none"),
        pytest.param(CustomCls.foo, id="slotwrapper"),  # type: ignore[attr-defined]
        pytest.param(lambda: None, id="lambda"),
        pytest.param(CustomCls.meth, id="func_nodoc"),
        pytest.param(CustomCls().meth, id="meth_nodoc"),
    ],
)
def test_skip(process_doc: ProcessDoc, obj: Callable[..., Any]) -> None:
    doc = inspect.getdoc(obj)
    assert process_doc(obj) == [doc or ""]


def test_defaults_simple(process_doc: ProcessDoc) -> None:
    def fn_test(
        s: str = "foo", n: None = None, i_: int = 1
    ) -> None:  # pragma: no cover
        r""":param s: Test S
        :param n: Test N
        :param i\_: Test I
        """  # noqa: D205
        del s, n, i_

    assert process_doc(fn_test) == [
        f":type s: {_escape_sat(':py:class:`str`')} (default: ``'foo'``)",
        ":param s: Test S",
        f":type n: {_escape_sat(':py:obj:`None`')} (default: ``None``)",
        ":param n: Test N",
        rf":type i\_: {_escape_sat(':py:class:`int`')} (default: ``1``)",
        r":param i\_: Test I",
        NONE_RTYPE,
    ]


def test_defaults_complex(process_doc: ProcessDoc) -> None:
    def fn_test(m: Mapping[str, int] = {}) -> None:  # pragma: no cover
        """:param m: Test M"""
        del m

    expected = (
        r":py:class:`~collections.abc.Mapping`\ \[:py:class:`str`, :py:class:`int`]"
    )
    assert process_doc(fn_test) == [
        f":type m: {_escape_sat(expected)} (default: ``{{}}``)",
        ":param m: Test M",
        NONE_RTYPE,
    ]


def test_qualname_overrides_class(app: Sphinx, testmod: ModuleType) -> None:
    assert testmod.Class.__module__ == "testmod"
    assert typehints_formatter(testmod.Class, app.config) == ":py:class:`~test.Class`"


def test_qualname_overrides_exception(app: Sphinx, testmod: ModuleType) -> None:
    assert testmod.Excep.__module__ == "testmod"
    assert typehints_formatter(testmod.Excep, app.config) == ":py:exc:`~test.Excep`"


# These guys aren’t listed as classes in Python’s intersphinx index:
@pytest.mark.parametrize(
    "annotation",
    [
        Any,
        AnyStr,
        NoReturn,
        Callable[[int], None],
        Union[int, str],
        Union[int, str, None],
    ],
    ids=lambda p: str(p).replace("typing.", ""),
)
def test_typing_classes(app: Sphinx, annotation: type) -> None:
    app.config.typehints_fully_qualified = True  # type: ignore[attr-defined]
    name = (
        getattr(annotation, "_name", None)
        or getattr(annotation, "__name__", None)
        or getattr(get_origin(annotation), "_name", None)
        # 3.6 _Any and _Union
        or annotation.__class__.__name__[1:]
    )
    output = typehints_formatter(annotation, app.config)
    assert output is None or output.startswith(f":py:data:`typing.{name}")


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires Python 3.10+")
def test_union_type(app: Sphinx) -> None:
    union = eval("int | str")  # noqa: S307, PGH001
    assert typehints_formatter(union, app.config) is None


@pytest.mark.parametrize(
    ("direc", "base", "sub"),
    [
        ("autoclass", "Class", "SubCl"),
        ("autoexception", "Excep", "Excep2"),
    ],
)
def test_autodoc(
    app: Sphinx,
    testmod: ModuleType,  # noqa: ARG001
    direc: str,
    base: str,
    sub: str,
) -> None:
    Path(app.srcdir, "index.rst").write_text(
        f"""\
.. {direc}:: testmod.{sub}
   :show-inheritance:
""",
    )
    app.build()
    out = Path(app.outdir, "index.html").read_text()
    assert not (ws := cast(StringIO, app._warning).getvalue()), ws  # noqa: SLF001
    assert re.search(
        r'<(code|span)?[^>]*><span class="pre">test\.</span></(code|span)>'
        f'<(code|span)?[^>]*><span class="pre">{sub}</span></(code|span)>',
        out,
    ), out
    assert f'<a class="headerlink" href="#test.{sub}"' in out, out
    assert re.search(rf"Bases: <code[^>]*><span[^>]*>(?:test\.)?{base}", out), out


def test_fwd_ref(app: Sphinx, make_module: Callable[[str, str], ModuleType]) -> None:
    make_module(
        "fwd_mod",
        """\
        from dataclasses import dataclass

        @dataclass
        class A:
           b: 'B'

        @dataclass
        class B:
            a: A
        """,
    )
    Path(app.srcdir, "index.rst").write_text(
        """\
.. autosummary::

   fwd_mod.A
   fwd_mod.B
""",
    )
    app.setup_extension("sphinx.ext.autosummary")

    app.build()

    out = Path(app.outdir, "index.html").read_text()
    buf = cast(StringIO, app._warning)  # noqa: SLF001
    warnings = [
        w
        for w in buf.getvalue().splitlines()
        if "Cannot treat a function defined as a local function" not in w
    ]
    assert not warnings, warnings
    # TODO(flying-sheep): actually reproduce
    # https://github.com/theislab/scanpydoc/issues/14
    assert "fwd_mod.A" in out, out


@pytest.mark.parametrize(
    "docstring",
    [
        """
        :param x: An x
        :returns: foo
                      A foo!
                  bar
                      A bar!
        """,
        """
        :returns: foo
                      A foo!
                  bar
                      A bar!
        :param x: An x
        """,
        """
        Parameters
        ----------
        x
            An x

        Returns
        -------
        foo
            A foo!
        bar
            A bar!
        """,
    ],
    ids=["last", "first", "numpydoc"],
)
@pytest.mark.parametrize(
    ("return_ann", "foo_rendered"),
    [
        pytest.param(tuple[str, int], ":py:class:`str`", id="tuple"),
        pytest.param(Optional[tuple[str, int]], ":py:class:`str`", id="tuple | None"),
        pytest.param(
            tuple[Mapping[str, float], int],
            r":py:class:`~collections.abc.Mapping`\ \["
            ":py:class:`str`, :py:class:`float`"
            "]",
            id="complex",
        ),
        pytest.param(Optional[int], None, id="int | None"),
    ],
)
def test_return_tuple(
    process_doc: ProcessDoc,
    docstring: str,
    return_ann: type,
    foo_rendered: str | None,
) -> None:
    is_numpydoc = "-----" in docstring

    def fn_test() -> None:  # pragma: no cover
        pass

    fn_test.__doc__ = docstring
    fn_test.__annotations__["return"] = return_ann
    lines = [
        l
        for l in process_doc(fn_test, run_napoleon=is_numpydoc)
        if l
        if not re.match(r"^:(rtype|param)( \w+)?:", l)
    ]
    if foo_rendered is None:
        assert lines == [
            ":returns: foo",
            "              A foo!",
            "          bar",
            "              A bar!",
        ]
    else:
        assert lines == [
            f":returns: foo : {foo_rendered}",
            "              A foo!",
            "          bar : :py:class:`int`",
            "              A bar!",
        ]


def test_return_nodoc(process_doc: ProcessDoc) -> None:
    def fn() -> tuple[int, str]:  # pragma: no cover
        """No return section."""
        return 1, ""

    res = process_doc(fn)
    assert len(res) == 3  # noqa: PLR2004
    assert res[0:2] == [inspect.getdoc(fn), ""]
    assert res[2].startswith(":rtype: :sphinx_autodoc_typehints_type:")


def test_load_error(make_app_setup: Callable[..., Sphinx]) -> None:
    with pytest.raises(
        ExtensionError,
        match=r"Can only use annotate_defaults.*when using sphinx-autodoc-typehints",
    ):
        make_app_setup(
            extensions=["sphinx.ext.autodoc", "scanpydoc.elegant_typehints"],
            annotate_defaults=True,
        )



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from sphinx.ext.napoleon import NumpyDocstring  # type: ignore[attr-defined]

__all__ = ["_parse_returns_section"]


def _parse_returns_section(self: NumpyDocstring, section: str) -> list[str]:  # noqa: ARG001
    lines_raw = list(self._dedent(self._consume_to_next_section()))
    if lines_raw[0] == ":":
        # Remove the “:” inserted by sphinx-autodoc-typehints
        # https://github.com/tox-dev/sphinx-autodoc-typehints/blob/a5c091f725da8374347802d54c16c3d38833d41c/src/sphinx_autodoc_typehints/patches.py#L66
        lines_raw.pop(0)
    lines = self._format_block(":returns: ", lines_raw)
    if lines and lines[-1]:
        lines.append("")
    return lines



"""Test things that aren’t sub-extension specific."""

from __future__ import annotations

import pkgutil
from typing import TYPE_CHECKING
from functools import partial
from importlib import import_module

import scanpydoc


if TYPE_CHECKING:
    from collections.abc import Callable

    import pytest
    from sphinx.application import Sphinx


DEPRECATED = frozenset({"scanpydoc.autosummary_generate_imported"})


def test_all_get_installed(
    monkeypatch: pytest.MonkeyPatch, make_app_setup: Callable[..., Sphinx]
) -> None:
    setups_seen: set[str] = set()
    setups_called: dict[str, Sphinx] = {}
    for _finder, mod_name, _ in pkgutil.walk_packages(
        scanpydoc.__path__, f"{scanpydoc.__name__}."
    ):
        mod = import_module(mod_name)
        if mod_name in DEPRECATED or not hasattr(mod, "setup"):
            continue
        setups_seen.add(mod_name)
        monkeypatch.setattr(mod, "setup", partial(setups_called.__setitem__, mod_name))

    app = make_app_setup()
    app.setup_extension("scanpydoc")

    assert set(setups_called) == setups_seen
    for app2 in setups_called.values():
        assert app is app2



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, cast, overload


class PyInfo(TypedDict):
    module: str
    fullname: str


class CInfo(TypedDict):
    """C / C++ info."""

    names: list[str]


class JSInfo(TypedDict):
    object: str
    fullname: str


if TYPE_CHECKING:
    from typing import Literal, TypeAlias

    Domain = Literal["py", "c", "cpp", "javascript"]
    DomainInfo: TypeAlias = PyInfo | CInfo | JSInfo


@overload
def linkcode_resolve(
    domain: Literal["py"], info: PyInfo
) -> str | None:  # pragma: no cover
    ...


@overload
def linkcode_resolve(
    domain: Literal["c", "cpp"], info: CInfo
) -> str | None:  # pragma: no cover
    ...


@overload
def linkcode_resolve(
    domain: Literal["javascript"], info: JSInfo
) -> str | None:  # pragma: no cover
    ...


def linkcode_resolve(domain: Domain, info: DomainInfo) -> str | None:
    from . import github_url

    if domain != "py":
        return None
    info = cast(PyInfo, info)
    if not info["module"]:
        return None
    return github_url(f'{info["module"]}.{info["fullname"]}')



"""This module exists just for rtd_github_links tests."""  # noqa: D404

from __future__ import annotations

from dataclasses import field, dataclass

from legacy_api_wrap import legacy_api


@dataclass
class TestDataCls:
    test_attr: dict[str, str] = field(default_factory=dict)


class TestCls:
    test_anno: int


def test_func() -> None:  # pragma: no cover
    pass


@legacy_api()
def test_func_wrap() -> None:  # pragma: no cover
    pass
